(*  analyze.ml takes the AST and does the following
  - Checks/infers types
  - Creates a scope list/symbol map(s)
  - Handles captures for lambdas
 *)

open Ast
open Strings

exception Name_error of string (* Identifier is undeclared or declared more than once *)
exception Type_error of string (* Something is of the wrong type *)
exception Access_error of string (* Something cannot be accessed or is in the wrong place *)
exception Internal_error (* Thrown when code that should not be reached is reached *)

(* Returns closest scope containing a symbol.
   Raises exception if symbol can't be found *)
let rec locateSymbol scopes scope sym =
  if StringMap.mem sym scope.symbols
  then scope
  else match scope.outer with
    None -> raise (Name_error ("Undeclared identifier: " ^ sym))
  | Some(outer) -> locateSymbol scopes (List.nth scopes outer) sym

(* Returns true if types are compatible *)
let rec compatible = function
    FloatT, IntT -> true (* Ints can be used as floats *)
  | ListT(ListT(l1)), ListT(ListT(l2)) -> compatible (ListT(l1), ListT(l2))
  | ListT(_), ListT(VoidT) -> true (* Empty list matches any list type *)
  | FigureT(None), FigureT(_) | FigureT(_), FigureT(None) -> true
  | FigDefT, _ | _, FigDefT -> false (* Figure definitions are never equal *)
  | t1, t2 -> if t1 = t2 then true else false

(* Adds symbol to scope, throws error if it's already there *)
let addSymbol sym typ scope = match sym, typ with
  "_animating", BoolT -> StringMap.add "_animating" BoolT scope
| "_visible", BoolT -> StringMap.add "_visible" BoolT scope
| "_animating", _ -> raise (Type_error "_animating must be a bool")
| "_visible", _ -> raise (Type_error "_visible must be a bool")
| sym, typ -> if StringMap.mem sym scope
    then raise (Name_error ("Identifier already declared: " ^ sym))
    else StringMap.add sym typ scope

(* Performs a function on a scope's symbols and returns updated scope list *)
let modSymbols func s t scopes ind =
  let list = func s t (List.nth scopes ind).symbols in
  let (_, l) = List.fold_left
    (fun (i, l) x -> (i+1, (
      if i = ind
      then {symbols=list; outer=x.outer; figure=x.figure; capture=x.capture; typ=x.typ}
      else x
    )::l)) (0, []) scopes
  in List.rev l

(* Add capture to a scope given scope list and index *)  
let addCapture sym loc scopes ind =
  let map = match (List.nth scopes ind).capture with
    None -> raise Internal_error | Some(capture) -> StringMap.add sym loc capture in
  let (_, l) = List.fold_left
    (fun (i, l) x -> (i+1, (
      if i = ind
      then {symbols=x.symbols; outer=x.outer; figure=x.figure; capture=Some(map); typ=x.typ}
      else x
    )::l)) (0, []) scopes
  in List.rev l
 
(* Add capture to function if necessary *) 
let capture scopes scope sym =
  if StringMap.mem sym (List.nth scopes scope).symbols
  then scopes
  else let rec locate sym scope =
    if StringMap.mem sym (List.nth scopes scope).symbols
    then scope
    else match (List.nth scopes scope).outer with
      None -> raise (Name_error ("Undeclared identifier: " ^ sym))
    | Some(outer) -> locate sym outer
    in addCapture sym (locate sym scope) scopes scope

(* Add declaration to figure scope *)  
let modFigureScope declType value scopes ind =
  let (_, l) = List.fold_left
    (fun (i, l) x -> (i+1, (
      match x.figure with None -> x | Some(fig) ->
      if i = ind
      then {symbols=x.symbols;
        outer=x.outer;
        figure = Some(match declType with
          "param" -> {name=fig.name; params=value::fig.params;
            rParams=fig.rParams; comps=fig.comps; vars=fig.vars}
        | "rParam" -> {name=fig.name; params=value::fig.params;
            rParams=value::fig.rParams; comps=fig.comps; vars=fig.vars}
        | "comp" -> {name=fig.name; params=fig.params;
            rParams=fig.rParams; comps=value::fig.comps; vars=fig.vars}
        | "var" -> {name=fig.name; params=fig.params;
            rParams=fig.rParams; comps=fig.comps; vars=value::fig.vars}
        | _ -> raise Internal_error);
        capture=x.capture; typ=x.typ}
      else x
    )::l)) (0, []) scopes
  in List.rev l

(* Return the type of arithmetic operation if valid *)
let aropType =
  let err t1 op t2 = raise (Type_error ("Cannot perform " ^ Strings.of_arop op ^
    " on " ^ Strings.of_type t1 ^ " and " ^ Strings.of_type t2)) in
  function
  IntT, Div, FloatT | FloatT, Div, IntT | FloatT, Div, FloatT | IntT, Div, IntT -> FloatT
| IntT, Divdiv, FloatT | FloatT, Divdiv, IntT | FloatT, Divdiv, FloatT | IntT, Divdiv, IntT -> IntT
| IntT, _, IntT -> IntT
| IntT, _, FloatT | FloatT, _, IntT -> FloatT | FloatT, _, FloatT -> FloatT
| PairT, _, IntT | PairT, _, FloatT | IntT, _, PairT
| FloatT, _, PairT | PairT, _, PairT -> PairT
| StringT, Plus, StringT -> StringT
| (ListT(_) as t1), Plus, (ListT(_) as t2) when t1 = t2 -> t1
| t1, op, t2 -> (err t1 op t2)

(* Return the type of logical operation if valid *)
let logopType = 
    let err t1 op t2 = raise (Type_error ("Cannot perform " ^ Strings.of_logop op ^
    " on " ^ Strings.of_type t1 ^ " and " ^ Strings.of_type t2)) in
  function
  BoolT, And, BoolT | BoolT, Or, BoolT -> BoolT
| t1, (And as op), t2 | t1, (Or as op), t2 -> (err t1 op t2)
| IntT, _, IntT | IntT, _, FloatT | FloatT, _, IntT | FloatT, _, FloatT -> BoolT
| (FigDefT as t1), op, (FigDefT as t2)
| (FigureT(_) as t1), op, (FigureT(_) as t2)
| (FuncT(_, _) as t1), op, (FuncT(_, _) as t2) -> (err t1 op t2)
| t1, Eqeq, t2 when t1 = t2 -> BoolT
| t1, Neq, t2 when t2 = t2 -> BoolT
| t1, op, t2 -> (err t1 op t2)

(* Return the type of unary operation if valid *)
let unopType =
    let err op t = raise (Type_error ("Cannot perform " ^ Strings.of_unop op ^
    " on " ^ Strings.of_type t)) in
  function
  Not, BoolT -> BoolT
| Neg, IntT -> IntT
| Neg, FloatT -> FloatT
| Question, (FigDefT as t)
| Question, (FigureT(_) as t)
| Question, (FuncT(_, _) as t) -> (err Question t)
| Question, _ -> BoolT
| op, t -> (err op t)
  
(* Add nth scope to scope list *)
let addScope scopes n outer figure capture typ =
  let scope = {symbols=StringMap.empty; outer=outer;
    figure=figure; capture=capture; typ=typ} in
  let (_, l) = List.fold_left
    (fun (i, l) x -> (i+1, if i = n then scope::l else x::l)) (0, []) scopes
  in List.rev l
  
(* Find scope given figure type *)
let figureScope figure scopes =
  let scopeList = List.map
    (fun x -> match x.figure with None -> None
    | Some(info) -> if figure = info.name then Some(x) else None) scopes
  in let scope = List.fold_left (fun s x -> match x with None -> s | Some(x) -> Some(x))
    None scopeList
  in match scope with
    None -> raise (Name_error ("Undeclared identifier: " ^ figure)) | Some(scope) -> scope
  
(* Creates a list with n empty scopes *)
let rec initScopeList l = function
  0 -> List.rev l
| n -> initScopeList
    ({symbols=StringMap.empty; outer=None; figure=None; capture=None; typ=None}::l) (n-1)

(* Determine the type of an expression
   Raise exception if type cannot be determined
   Raise exception if types are invalid *)
let rec exprType scope scopes expr =
  let rec exprType_ = function
    Int(_) -> IntT
  | Float(_) -> FloatT
  | String(_) -> StringT
  | Bool(_) -> BoolT
  | Function(typ, args, (scope, stmts)) -> FuncT(typ, List.map snd args)
  | List(l) -> let rec listType = function
      [] -> VoidT
    | l -> match List.fold_right (fun el t ->
          match el with List(_) -> t | expr -> exprType_ el) l VoidT with
        VoidT -> ListT(listType (List.fold_left (fun l el ->
          match el with List(n) -> l @ n | _ -> l) [] l))
      | t -> t
    in ListT(List.fold_left
      (fun t e -> if compatible (t, exprType_ e)
        then t else raise (Type_error ("Unexpected " ^ Strings.of_type (exprType_ e) ^
          " in list of " ^ Strings.of_type t)))
    (listType l) l)
  | Color(_) -> ColorT
  | Pair(x, y) -> (match ((exprType_ x), (exprType_ y)) with
    IntT, IntT -> PairT
    | IntT, FloatT -> PairT
    | FloatT, IntT -> PairT
    | FloatT, FloatT -> PairT
    | _, _ -> raise (Type_error "Pair type must have int or float parts"))
  | Subscript(list, index) ->
    (match ((exprType_ list), (exprType_ index)) with
      ListT(typ), IntT -> typ
    | typ, _ -> raise (Type_error ("Cannot subscript " ^ Strings.of_type typ)))
  | Dot(figure, id) -> (match (exprType_ figure) with
      FigureT(Some(fig)) -> let figScope = figureScope fig scopes in
        if match figScope.figure with None -> raise Internal_error
        | Some(fig) -> List.mem id fig.vars
        then raise (Access_error ("Cannot access private var " ^ id))
        else (try StringMap.find id figScope.symbols
      with Not_found -> raise (Name_error ("Undeclared identifier: " ^ id)))
    | FigureT(None) -> (match id with
          "_visible" | "_animating" -> BoolT
        | _ -> raise (Access_error ("Cannot access " ^ id ^ " from generic figure type")))
    | typ -> raise (Type_error ("Dot operator cannot be used on "
      ^ Strings.of_type typ)))
  | Id(id) -> StringMap.find id (locateSymbol scopes scope id).symbols
  | Arop(e1, op, e2) -> aropType ((exprType_ e1), op, (exprType_ e2))
  | Logop(e1, op, e2) -> logopType ((exprType_ e1), op, (exprType_ e2))
  | Unop(op, expr) -> unopType (op, (exprType_ expr))
  | Call(expr, args) ->
      (match exprType_ expr with
        FuncT(ret, argTypes) ->
          if argTypes = List.map (fun x -> exprType_ x) args
          then ret
          else raise (Type_error ("Arguments of function not of proper types"))
      | FigDefT ->
          if List.length args <> 1 then raise (Type_error "Figure casts take exactly one argument")
          else (match (expr, exprType_ (List.nth args 0)) with
            Id(id), FigureT(None) -> FigureT(Some(id))
          | _, t -> raise (Type_error "Figure casts can only be done on generic figure types"))
      | typ -> raise (Type_error ("Cannot make call to " ^ Strings.of_type typ)))
  | FigInstance(id, params) -> let figScope = (figureScope id scopes) in
      let types = List.map
        (fun (id, value) -> StringMap.find id figScope.symbols, (exprType_ value)) params in
      List.iter (fun (expected, typ) ->
        if compatible (expected, typ) then ()
        else raise (Type_error ("Expected " ^ Strings.of_type expected ^
        ", instead found " ^ Strings.of_type typ))) types ;
      match figScope.figure with None -> raise Internal_error | Some(fig) ->
      List.iter (fun x -> if List.mem x (fst (List.split params)) then ()
        else raise (Type_error ("Missing required parameter " ^ x ^
        " in instance of " ^ id))) fig.rParams ;
      FigureT(Some(id))
  in exprType_ expr
  
(* Returns true if the expression represents a component.
   Used to make sure components aren't directly assigned *)
let isComp scope scopes = function
  Dot(figure, id) -> (match (exprType scope scopes figure) with
      FigureT(Some(fig)) -> let figScope = (figureScope fig scopes) in
        (match figScope.figure with None -> raise Internal_error | Some(figInfo) ->
        if List.mem id figInfo.comps then true else false)
    | _ -> false)
| Id(id) -> (match (locateSymbol scopes scope id).figure with
      None -> false
    | Some(figInfo) -> if List.mem id figInfo.comps then true else false)
| _ -> false

(* Extract leftMost id from assignable expression.
   Throws error if expression is not assignable.
   Used for handling captures *)
let rec extractId = function
  Id(id) -> id
| Dot(Id(id), _) | Subscript(Id(id), _) -> id
| Dot(expr, _) | Subscript(expr, _) -> extractId expr
| _ -> raise (Type_error "Expression is not assignable")

(* Perform captures for a function *)
let makeCaptures scopes current stmts =
  let rec makeCapturesExpr scopes = function
    Id(id) -> capture scopes current id
  | Dot(expr, _) | Subscript(expr, _) -> capture scopes current (extractId expr)
  | Pair(expr1, expr2) | Arop(expr1, _, expr2) | Logop(expr1, _, expr2) ->
      makeCapturesExpr (makeCapturesExpr scopes expr2) expr1
  | Unop(_, expr) -> makeCapturesExpr scopes expr
  | List(exprs) | Call(_, exprs) ->
      List.fold_left (fun scopes e -> makeCapturesExpr scopes e) scopes exprs
  | FigInstance(_, list) ->
      List.fold_left (fun scopes (_, e) -> makeCapturesExpr scopes e) scopes list
  | _ -> scopes
  in let rec _makeCaptures scopes = function
    [] -> scopes
  | Expr(expr)::tl -> _makeCaptures (makeCapturesExpr scopes expr) tl
  | Assign(expr, value)::tl -> let scopes = capture scopes current (extractId expr) in
      _makeCaptures (makeCapturesExpr scopes value) tl
  | Return(Some(expr))::tl -> _makeCaptures (makeCapturesExpr scopes expr) tl
  | If(expr, (_, stmts1), (_, stmts2))::tl ->
      let scopes = _makeCaptures (_makeCaptures scopes stmts1) stmts2 in
      _makeCaptures (makeCapturesExpr scopes expr) tl
  | While(expr, (_, stmts))::tl | DoWhile((_, stmts), expr)::tl ->
    _makeCaptures (makeCapturesExpr (_makeCaptures scopes stmts) expr) tl
  | _::tl -> _makeCaptures scopes tl
in _makeCaptures scopes stmts

(* Make the scope list/symbol tables
   This is the main function of analyze.ml
   It takes one argument, an Ast.program
   It returns a list of scopes (scope type defined in ast.ml) *)
let makeTable sList =
  let rec makeScope scopes current = function
    [] -> scopes
  | Figure(index, id, decls)::tl -> (* Figure definition *)
      if current <> 0 then raise (Access_error "Figure definitions must be in the global scope")
      else let scopes = modSymbols addSymbol id FigDefT scopes current in
      let scopes = addScope scopes index (Some current)
        (Some {name=id; params=[]; rParams=[]; comps=[]; vars=[]}) None None in
      let decls = [(Param, "_visible", Some(BoolT), None);
        (Param, "_animating", Some(BoolT), None)] @ decls in
      makeScope (List.fold_left (fun scopes decl -> let scopes = match decl with
          Param, "_animating", _, _ | Param, "_visible", _, _ -> scopes
        | Var, "_update", Some(FuncT(VoidT, [IntT])), _ -> scopes
        | Var, "_update", None, Some(expr)
            when (exprType (List.nth scopes current) scopes expr) = FuncT(VoidT, [IntT]) -> scopes
        | Var, "_update", _, _ -> raise (Type_error "_update must have type (int)")
        | _, "_animating", _, _ -> raise (Type_error "_animating must be declared as a param")
        | _, "_visible", _, _ -> raise (Type_error "_visible must be declared as a param")
        | _, "_update", _, _ -> raise (Type_error "_update must be declared as a var")
        | Param, sym, Some(_), None -> modFigureScope "rParam" sym scopes index
        | Param, sym, None, Some(_) -> modFigureScope "param" sym scopes index
        | Comp, sym, _, _ -> modFigureScope "comp" sym scopes index
        | Var, sym, _, _ -> modFigureScope "var" sym scopes index
        | _ -> scopes in
        parseDecl scopes index decl) scopes decls) current tl
  | Decl(decls)::tl -> let scopes = List.fold_left
    (fun scopes decl -> parseDecl scopes current decl) scopes decls
    in makeScope scopes current tl
  | Expr(expr)::tl -> ignore (exprType (List.nth scopes current) scopes expr) ;
      makeScope scopes current tl
  | Assign(expr, value)::tl ->
      (match expr with
        Id(_) | Dot(_, _) | Subscript(_, _) ->
          let t1 = exprType (List.nth scopes current) scopes expr
          in let t2 = exprType (List.nth scopes current) scopes value
          in if isComp (List.nth scopes current) scopes expr 
          then raise (Access_error ("Cannot assign components directly"))
          else if compatible (t1, t2)
          then makeScope scopes current tl
          else raise (Type_error ("Cannot assign " ^ Strings.of_type t2 ^
            " to " ^ Strings.of_type t1))
      | _ -> raise (Type_error "Expression is not assignable"))
  | If(expr, (ind1, stmts1), (ind2, stmts2))::tl ->
      ignore (exprType (List.nth scopes current) scopes expr) ;
      let scopes = makeScope (addScope scopes ind1 (Some current) None None None) ind1 stmts1
      in let scopes = makeScope (addScope scopes ind2 (Some current) None None None) ind2 stmts2
      in makeScope scopes current tl
  | DoWhile ((ind, stmts), expr)::tl
  | While(expr, (ind, stmts))::tl ->
      ignore (exprType (List.nth scopes current) scopes expr) ;
      let scopes = makeScope (addScope scopes ind (Some current) None None None) ind stmts
      in makeScope scopes current tl
  | Return(e)::tl ->
      (* Returns type of current function *)
      let rec funcType scope scopes = match (List.nth scopes scope).typ with
        Some(typ) -> typ
      | None -> match (List.nth scopes scope).outer with
          None -> raise (Access_error "Return statements must be inside functions")
        | Some(outer) -> funcType outer scopes in
      let t = funcType current scopes in match e, t with
        Some(Function(rt, args, (fIndex, stmts))), t
          when (exprType (List.nth scopes current) scopes (Function(rt, args, (fIndex, stmts)))) = t ->
            let scopes = makeFunctionScope scopes current rt args fIndex stmts in
            makeScope scopes current tl
      | Some(e), t when (exprType (List.nth scopes current) scopes e) = t -> makeScope scopes current tl
      | None, VoidT -> makeScope scopes current tl
      | None, t -> raise (Type_error ("Function must return " ^ Strings.of_type t ^ " not void"))
      | Some(e), t -> raise (Type_error ("Function must return " ^ Strings.of_type t ^ " not " ^
          Strings.of_type (exprType (List.nth scopes current) scopes e)))
  and parseDecl scopes index = function
    Comp, sym, None, Some(expr) ->
    (match exprType (List.nth scopes index) scopes expr with
      FigureT(f) -> modSymbols addSymbol sym (FigureT f) scopes index
    | _ -> raise (Type_error ("Component " ^ sym ^
      " must be of figure type.")))
  | Comp, sym, Some(FigureT(f)), None -> modSymbols addSymbol sym (FigureT f) scopes index
  | Comp, sym, _, None -> raise (Type_error ("Component " ^ sym ^
      " must be of figure type."))
  | _, sym, Some(typ), None -> modSymbols addSymbol sym typ scopes index
  | _, sym, None, Some(expr) ->
    let typ = (exprType (List.nth scopes index) scopes expr)
    in let rec isEmpty = function VoidT -> true | ListT(l) -> isEmpty l | _ -> false
    in let scopes = if isEmpty typ then raise (Type_error "Cannot infer type of empty list")
      else modSymbols addSymbol sym typ scopes index
    in (match expr with
      Function(t, args, (fIndex, stmts)) -> makeFunctionScope scopes index t args fIndex stmts
    | _ -> scopes)
  | _, sym, _, _ -> raise Internal_error
  and makeFunctionScope scopes index t args fIndex stmts =
    let scopes = List.fold_left
      (fun scopes (sym, typ) -> modSymbols addSymbol sym typ scopes fIndex)
      (addScope scopes fIndex (Some index) None (Some StringMap.empty) (Some t)) args
    in let scopes = makeScope scopes fIndex stmts
    in makeCaptures scopes fIndex stmts
  in makeScope (initScopeList [] (!Ast.numScopes + 1)) 0 sList