%{ open Ast %}

%token <string> STRING ID
%token <int> INT
%token <float> FLOAT
%token <string * string * string * float> COLOR
%token PLUS MINUS TIMES DIV DIVDIV EXP PERCENT PLUSEQ MINUSEQ TIMESEQ DIVEQ DIVDIVEQ EXPEQ PERCENTEQ 
%token EQ PLUSEQ MINUSEQ TIMESEQ DIVEQ DIVDIVEQ EXPEQ PERCENTEQ
%token EQEQ NEQ GT LT GTE LTE NOT AND OR QUESTION
%token DOT COLON COMMA LPAREN RPAREN LBRACKET RBRACKET LBRACE RBRACE LANGLE RANGLE
%token PARAM COMP VAR IF ELSEIF ELSE DO WHILE TRUE FALSE RETURN
%token FIGURE T_INT T_FLOAT T_STRING FUNC BOOL T_COLOR PAIR
%token NL EOF

%nonassoc NL
%left COMMA
%right EQ PLUSEQ MINUSEQ TIMESEQ DIVEQ DIVDIVEQ EXPEQ PERCENTEQ
%nonassoc EQEQ NEQ
%nonassoc LT GT LTE GTE
%left OR
%left AND
%left PLUS MINUS
%left TIMES DIV DIVDIV PERCENT
%left EXP
%nonassoc UMINUS NOT QUESTION
%left DOT
%nonassoc LPAREN RPAREN LBRACKET RBRACKET LBRACE RBRACE LANGLE RANGLE

%start program
%type <Ast.program> program

%%

program:
    program_contents { List.rev $1 }

program_contents:
    ws { [] }
  | program_contents stmt NL { $2 :: $1 }
  | program_contents stmt EOF { $2 :: $1 }
 
ws:
    /* nothing */ {}
  | NL {}
  
figure:
    FIGURE ID ws LBRACE figure_body figure_decl ws RBRACE { Figure(Ast.nextScope(), $2, $5 @ $6) }
  | FIGURE ID ws LBRACE ws RBRACE { Figure(0, $2, []) }
  
block:
    LBRACE RBRACE { Ast.nextScope(), [] }
  | LBRACE stmt_list stmt ws RBRACE { Ast.nextScope(), List.rev ($3 :: $2) }
  
typ:
    T_INT { IntT }
  | T_FLOAT { FloatT }
  | T_STRING { StringT }
  | BOOL { BoolT }
  | T_COLOR { ColorT }
  | PAIR { PairT }
  | ID LBRACE RBRACE { FigureT(Some($1)) }
  | FIGURE LBRACE RBRACE { FigureT(None) }
  | typ LBRACKET RBRACKET { ListT($1) }
  | typ LPAREN typ_list_opt RPAREN { FuncT($1, $3) }
  | LPAREN typ_list_opt RPAREN { FuncT(VoidT, $2) }
  
typ_list:
    typ { [$1] }
  | typ_list COMMA ws typ { $4 :: $1 }
  
typ_list_opt:
    /* nothing */ { [] }
  | typ_list { List.rev $1 }
  
stmt_list:
    /* nothing */ { [] }
  | stmt_list stmt NL { $2 :: $1 }
  
figure_decl:
    PARAM ws decl_list { List.map (fun (x, y, z) -> (Param, x, y, z)) (List.rev $3) }
  | COMP ws decl_list { List.map (fun (x, y, z) -> (Comp, x, y, z)) (List.rev $3) }
  | VAR ws decl_list { List.map (fun (x, y, z) -> (Var, x, y, z)) (List.rev $3) }
  
figure_body:
    /* nothing */ { [] }
  | figure_body figure_decl NL { $1 @ $2 }
  
decl_list:
    decl { [$1] }
  | decl_list COMMA ws decl { $4 :: $1 }
  
decl:
    ID EQ typ { $1, Some($3), None }
  | ID EQ expr { $1, None, Some($3) }
  
arg_list:
    ID COLON typ { [$1, $3] }
  | arg_list COMMA ws ID COLON typ { ($4, $6) :: $1 }

arg_list_opt:
    /* nothing */ { [] }
  | arg_list { List.rev $1 }
  
expr_list:
    expr { [$1] }
  | expr_list COMMA ws expr { $4 :: $1 }

expr_list_opt:
    /* nothing */ { [] }
  | expr_list { List.rev $1 }
  
param:
    ID COLON expr { $1, $3 }

param_list:
    param { [$1] }
  | param_list ws param { $3 :: $1 }

param_list_opt:
    COLON { [] }
  | param_list { List.rev $1 }
  
elseif:
    /* nothing */ { Ast.nextScope(), [] }
  | ELSE ws block { $3 }
  | ELSEIF expr ws block elseif { Ast.nextScope(), [If($2, $4, $5)] }

stmt:
    expr { Expr($1) }
  | figure { $1 }
  | VAR ws decl_list { Decl(List.map (fun (x, y, z) -> (Var, x, y, z)) (List.rev $3)) }
  | expr EQ expr { Assign($1, $3) }
  | expr PLUSEQ expr { Assign($1, Arop($1, Plus, $3)) }
  | expr MINUSEQ expr { Assign($1, Arop($1, Minus, $3)) }
  | expr TIMESEQ expr { Assign($1, Arop($1, Times, $3)) }
  | expr DIVEQ expr { Assign($1, Arop($1, Div, $3)) }
  | expr DIVDIVEQ expr { Assign($1, Arop($1, Divdiv, $3)) }
  | expr EXPEQ expr { Assign($1, Arop($1, Exp, $3)) }
  | expr PERCENTEQ expr { Assign($1, Arop($1, Percent, $3)) }
  | RETURN expr { Return(Some($2)) }
  | RETURN { Return(None) }
  | WHILE expr ws block { While($2, $4) }
  | DO ws block ws WHILE expr { DoWhile($3, $6) }
  | IF expr ws block elseif { If($2, $4, $5) }
    
expr:
    LPAREN expr RPAREN { $2 }
  | INT { Int($1) }
  | STRING { String($1) }
  | FLOAT { Float($1) }
  | TRUE { Bool(true) }
  | FALSE { Bool(false) }
  | COLOR { Color($1) }
  | LANGLE expr COMMA ws expr RANGLE { Pair($2, $5) }
  | ID { Id($1) }
  | FUNC LBRACKET typ RBRACKET LPAREN arg_list_opt RPAREN block { Function($3, $6, $8) }
  | FUNC LPAREN arg_list_opt RPAREN block { Function(VoidT, $3, $5) }
  | FUNC LBRACKET typ RBRACKET LPAREN arg_list_opt RPAREN expr { Function($3, $6, (Ast.nextScope(), [Return(Some($8))])) }
  | LBRACKET expr_list_opt RBRACKET { List($2) }
  | expr LBRACKET expr RBRACKET { Subscript($1, $3) }
  | expr DOT ID { Dot($1, $3) }
  | expr PLUS expr { Arop($1, Plus, $3) }
  | expr MINUS expr { Arop($1, Minus, $3) }
  | expr TIMES expr { Arop($1, Times, $3) }
  | expr DIV expr { Arop($1, Div, $3) }
  | expr DIVDIV expr { Arop($1, Divdiv, $3) }
  | expr EXP expr { Arop($1, Exp, $3) }
  | expr PERCENT expr { Arop($1, Percent, $3) }
  | expr EQEQ expr { Logop($1, Eqeq, $3) }
  | expr NEQ expr { Logop($1, Neq, $3) }
  | expr GT expr { Logop($1, Gt, $3) }
  | expr LT expr { Logop($1, Lt, $3) }
  | expr GTE expr { Logop($1, Gte, $3) }
  | expr LTE expr { Logop($1, Lte, $3) }
  | expr AND expr { Logop($1, And, $3) }
  | expr OR expr { Logop($1, Or, $3) }
  | NOT expr { Unop(Not, $2) }
  | QUESTION expr { Unop(Question, $2) }
  | MINUS expr %prec UMINUS { Unop(Neg, $2) }
  | func { $1 }
  | ID LPAREN param_list_opt RPAREN { FigInstance($1, $3) }

func:
    ID LPAREN expr_list_opt RPAREN { Call(Id($1), $3) }
  | expr DOT ID LPAREN expr_list_opt RPAREN { Call(Dot($1, $3), $5) }
  | func LPAREN expr_list_opt RPAREN { Call($1, $3) }
  | LPAREN expr RPAREN LPAREN expr_list_opt RPAREN { Call($2, $5) }
