(* translate.ml takes the AST and a scopes list and is responsible for
    java code generation *)

open Ast
open Strings
open Analyze
open Printf 
open Sys
open Programname

exception Internal_error of string

(* write a string to file *)
let dir = Programname.get_name ^ "_java/"

let write str file = 
    if (Sys.file_exists (dir ^ "GeneratedCode/" ^ file))
	then
	()
	else
    let oc = open_out (dir ^ "GeneratedCode/" ^file) in
	fprintf oc "%s\n" str;
	close_out oc 


(* list of captures *)
let captures = ref []
(* list of stdlib *)
let stdlib = ref []

let header = "package GeneratedCode;\n\nimport Stdlib.*;\nimport Type.*;\nimport Main.*;\nimport BuiltIns.*;\n\n"
let seperator = "-----------------------------------------------"

(* prefix on all user-defined identifiers *)
let id_prefix = ""

(* for unnamed functions *)
let arbitrary_function_counter = ref (-1)

(* formats a string list *)
let tr_list l = List.fold_left (fun s x -> s ^ x ^
  (let c = x.[(String.length x) - 1] in 
  	if c = ';' || c = '}' || c = '\n'
  	then ""
    else ";\n")) "" l

let tr_list2 = function
      [] -> ""
    | hd::tl -> (List.fold_left (fun s x -> s ^ ", " ^ x) hd tl)

let tr_list3 l = List.fold_left (fun s x -> s ^ x) "" l

let helper = (fun a b d -> (match d with
		    (a2, b2) -> (a2 = a) || (b2 = b)))

let helper3 = fun s x -> match s with
	(a, b) -> match x with
		    ((c, d),e) -> (a=c)&&(b=d)

let is_captured search_cap = List.exists (helper3 search_cap) !captures
let in_stdlib symbol = List.exists (fun a -> a = symbol) !stdlib

(* finds the index of a symbol *)
let rec scope_of_symbol symbol cur_scope scopes = 
    let scope = (try List.nth scopes cur_scope with Invalid_argument(str)
	-> raise (Internal_error ("Scope index out of bounds.")))
    
    
    in

    if (StringMap.exists (fun a b -> a = symbol) scope.symbols) then
	cur_scope
    else
	match scope.capture with
	  None ->  (match scope.outer with
			  None -> raise (Internal_error ("Scope not found")) 
			| Some(num) -> scope_of_symbol symbol num scopes)
	| Some(map) -> 
	    try (StringMap.find symbol map) with 
	    Not_found -> (match scope.outer with
			  None -> raise (Internal_error("Scope not found"))
			| Some(num) -> scope_of_symbol symbol num scopes)

(* finds the type of a symbol *)
let rec type_of_symbol symbol cur scopes = 
	let containingScope = (
	    try List.nth scopes cur with Invalid_argument(str) ->
		raise (Internal_error ("Scope not found in type_of_symbol"))) in
	    try (
		(StringMap.find symbol containingScope.symbols) 
	    ) with Not_found -> match containingScope.outer with
		  None -> raise Not_found 
		| Some(num) -> type_of_symbol symbol num scopes

(* ignores the next expr *)
let ignore_expr expr = ""
		
(* am i parsing the std lib? *)
let stdlib_parse = ref true

(*captures helper class*)
let captures_helper search_cap =
		let index = 
		    match (List.find (helper3 search_cap) !captures)
		    with (a, b) -> b
		in
		"Captures.record[" ^ (string_of_int index) ^ "]"

    
(* global record of figures *)
let figure_record = ref StringMap.empty

(* translate a typ *)
let rec tr_typ = function
      IntT -> "IntType"
    | FloatT -> "FloatType"
    | StringT -> "StringType"
    | BoolT -> "BoolType"
    | ColorT -> "ColorType"
    | PairT -> "PairType"
    | FigureT(arg) -> (match arg with
			  None -> "Figure"
			| Some(str) -> str)
    | VoidT -> "void"
    | FigDefT -> "FigDefT"
    | ListT(typ) -> "ListType<" ^ (tr_typ typ)  ^ ">"
    | FuncT (typ, typ_list) -> "FunctionType"

let rec type_of_expr expr cur scopes = match expr with
      Id(str) -> tr_typ (type_of_symbol str cur scopes)
    | Int(_) -> "IntType"
    | Float(_) -> "FloatType"
    | String(_) -> "StringType"
    | Bool(_) -> "BoolType"
    | Function(_,_,_) -> "FunctionType"
    | List(_) -> "ListType"
    | Color(_) -> "ColorType"
    | Pair(_,_) -> "PairType"
    | FigInstance(_,_) -> "FigInstance"
    | Subscript(_,_) -> "Subscript"
    | Dot(_,_) -> "Dot"
    | Arop(_,_,_) -> "Arop"
    | Logop (_, _, _) -> "Logop"
    | Unop (_, _) -> "Unop"
    | Call(_, _) -> "Call"

let rec type_of_expr2 expr cur scopes = match expr with
      Id(str) -> "Id"
    | Int(_) -> "IntType"
    | Float(_) -> "FloatType"
    | String(_) -> "StringType"
    | Bool(_) -> "BoolType"
    | Function(_,_,_) -> "FunctionType"
    | List(_) -> "ListType"
    | Color(_) -> "ColorType"
    | Pair(_,_) -> "PairType"
    | FigInstance(_,_) -> "FigInstance"
    | Subscript(_,_) -> "Subscript"
    | Dot(_,_) -> "Dot"
    | Arop(_,_,_) -> "Arop"
    | Logop (_, _, _) -> "Logop"
    | Unop (_, _) -> "Unop"
    | Call(_, _) -> "Call"


(* translate an expression into java *)
let rec tr_expr expr scopes cur = 

	let tr_arop arop arg1 arg2 sc cur = match arop with 
	      Plus -> "Arop.fogl_plus(" ^ 
		(tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Minus -> "Arop.fogl_minus(" 
		    ^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Times -> "Arop.fogl_times(" 
		    ^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Div -> "Arop.fogl_div(" 
		    ^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Divdiv -> "Arop.fogl_divdiv(" 
		    ^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Exp -> "Arop.fogl_exp(" 
		    ^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Percent -> "Arop.fogl_percent(" 
		    ^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	in
	let tr_logop logop arg1 arg2 sc cur = match logop with
	      Eqeq -> "Arop.fogl_eqeq(" 
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Neq -> "Arop.fogl_neq("
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Gt -> "Arop.fogl_gt("
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Lt -> "Arop.fogl_lt(" 
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Gte -> "Arop.fogl_gte("
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Lte -> "Arop.fogl_lte("
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | And -> "Arop.fogl_and("
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	    | Or -> "Arop.fogl_or("
		^ (tr_expr arg1 sc cur) ^ ", " ^ (tr_expr arg2 sc cur) ^ ")"
	in
	let tr_unop unop arg sc cur = match unop with
	      Neg -> "Arop.fogl_neg(" ^ (tr_expr arg sc cur) ^ ")"
	    | Not -> "Arop.fogl_not(" ^ (tr_expr arg sc cur) ^ ")"
	    | Question -> "Arop.fogl_question(" ^ (tr_expr arg sc cur) ^ ")"
	in

	let tr_figinstance str stuff scopes cur =
	    let fig_cons = StringMap.find str !figure_record
	    in
	    
	    (*
	    (let liststuff = List.map (fun x -> match x with
				    (a,b) -> a ^ " " ^ (tr_expr b scopes )  )
			    stuff
	    in
	    
	    let constructors = tr_list2 liststuff in
	    *)
	    let hef = fun x y -> match y with
				    (c, d) -> c = x
	    in
	    
	    let elt = fun x -> (List.find (hef x) stuff) 
	    in
	    let constructors = List.fold_left (fun a b -> a ^ (if a = "" then "" else ", ") ^ (
		try
		(match (elt b) with
		    (xx, yy) -> tr_expr yy scopes cur) with 
		Not_found -> "null")) 
		
		"" (List.rev fig_cons)
	    in
	    if (constructors = "null" ) then
		"new " ^ str ^ "((AbstractType)" ^ constructors ^ ")" 
	    else
		"new " ^ str ^ "(" ^ constructors ^ ")" 
	    
	in
    match expr with 
	  Int(arg) -> "new IntType(" ^ (string_of_int arg) ^ ")"
	| Float(arg) -> "new FloatType(" ^ (string_of_float arg) ^ ")"
	| String(arg) -> "new StringType(\"" ^ arg ^ "\")" 
	| Bool(arg) -> "new BoolType(" ^ (string_of_bool arg) ^ ")"
	| Function(a, b, c) -> "THIS SHOULD NEVER BE CALLED."
	(*
		let func_name =
		    arbitrary_function_counter :=
			!arbitrary_function_counter + 1;
		    "Func_" ^ (string_of_int cur) ^ "_" ^
			(string_of_int !arbitrary_function_counter)
		in
		let tr_arglist arglist = 
		    let tr_s x = match x with
		    (str, typ) -> (tr_typ typ) ^ " " ^ str
		in
		    List.map (fun x -> tr_s x) arglist   
		    in
		let super_str = 
		    (header ^"public class " ^
			func_name ^ " extends FunctionType {\n" 
			^ (tr_typ a) ^ " call(" ^ 
			(tr_list2 (tr_arglist b)) ^ ") {\n" 
			^ (tr_block c scopes cur)
			^ "}\n}\n"  
		    )
		in
		write super_str ("Func_" ^ str ^ ".java");
		" new " ^ func_name ^ "()"
*)
	| List(arg) -> "new ListType(" ^ 
	    (List.fold_left 
		(fun str x -> if (str = "") then 
		    str ^ (tr_expr x scopes cur) else 
		    str ^ ", " ^ (tr_expr x scopes cur)) "" arg) 
	    ^ ")"
	| Color(arg) -> (match arg with
	     (s1, s2, s3, fl) -> ("new ColorType(0x" ^ 
		s1 ^ ", 0x" ^ s2 ^ ", 0x" ^ s3 ^
		", " ^ (string_of_float fl)^ ")") )
	| Pair(arg1, arg2) -> "new PairType(" ^ (tr_expr arg1 scopes cur) ^ ", " 
				^ (tr_expr arg2 scopes cur) ^ ")"
	| Subscript(arg1, arg2) -> "(" ^ (tr_expr arg1 scopes cur) 
		^ ").get(" ^ (tr_expr arg2 scopes cur ) ^ ")"
	| Dot(arg1, arg2) -> "(" ^ (tr_expr arg1 scopes cur) ^ ")." ^ arg2 ^ ""
	| Id(arg1) -> 
	    (* is_captured arg1, cur is not returning true *)
	    let arg_scope_index =
		scope_of_symbol arg1 cur scopes
	    in
	    if arg_scope_index = 9001 
		then (print_endline "OVER 9000 ERROR";"") else		
	    if (is_captured (arg1, arg_scope_index)) then
		"(("^ (tr_typ (type_of_symbol arg1 cur scopes)) ^")"^
		captures_helper (arg1, arg_scope_index) ^ ")"
	    else 
		if ((in_stdlib arg1)&&(arg_scope_index = 0)) then
		    ("Stdlib.Lib." ^ arg1)
		else
		    (id_prefix ^ arg1) 
	| Arop(arg1, arop, arg2) -> tr_arop arop arg1 arg2 scopes cur
	| Logop(arg1, logop, arg2) -> tr_logop logop arg1 arg2 scopes cur
	| Unop(unop,arg) -> tr_unop unop arg scopes cur
	| Call(arg1, arg2) -> 
	    let arglist = List.map (fun x -> tr_expr x scopes cur) arg2 in
	    (tr_expr arg1 scopes cur) ^ ".call(" ^ (tr_list2 arglist) ^")"
	| FigInstance(arg1, arg2) -> tr_figinstance arg1 arg2 scopes cur

(* BEWARE! THE FOLLOWING IS ALMOST IMPOSSIBLE TO READ. *)

(* translate a statement into java *)
let rec tr_stmt stmt scopes cur = 

    let tr_block block scopes cur= 
	(let almost = 
	    (match block with
		(arg1, arg2) -> List.map 
		(fun x -> tr_stmt x scopes arg1) arg2)
	in tr_list almost)
    in

(* translate a list of decl *)
let tr_decls decls scopes cur = 

    let tr_typ_opt arg str scopes cur = match arg with
	  None -> let containingScope = (
	    try List.nth scopes cur with Invalid_argument(str) -> raise
		(Internal_error ("Scope not found in tr_typ_opt"))) (* this is where you search teh scopes list for the type*) 
	  
	    in
	    tr_typ (StringMap.find str containingScope.symbols) 
	| Some(typ) -> tr_typ typ
    in

    let tr_expr_opt arg scopes cur = match arg with
	  None -> ""
	| Some(expr) -> " = " ^ (tr_expr expr scopes cur)

    in
   let tr_decl decl scopes cur = (match decl with
	(decl_typ, str, typ_opt, expr_opt) ->

	if is_captured (str, cur) then
	    let other = (tr_expr_opt expr_opt scopes cur) in
	    if other = "" then "" else
		   (captures_helper (str, cur) ^
		   other)
	else
	(
	match decl_typ with 
    	  Var ->   
	    (match expr_opt with
	      None ->  
		("" ^ (tr_typ_opt typ_opt str scopes cur) ^ 
		" " ^ id_prefix ^ str ^ (tr_expr_opt expr_opt scopes cur)) ^
		(match typ_opt with Some(ListT(_)) -> " = new ListType()" | _ -> "")
	    | Some(expr) ->   
	    (match expr with
	      Function(a, b, c) -> 
		let tr_arglist arglist = 
		    let tr_s x = match x with
		    (str, typ) -> (tr_typ typ) ^ " " ^ str
		in
		    List.map (fun x -> tr_s x) arglist   
		    in
	    
		(match cur with
		0 ->
		(* function with no scope *)
		let super_str = 
		    (header ^"class Func_" ^ 
			str ^ " extends FunctionType {\n"
			^ "public static Func_" ^ str ^ " " ^ str ^ " = new Func_" ^ str ^ "();\n"
			^ "public " ^ (tr_typ a) ^ " call(" ^ 
			(tr_list2 (tr_arglist b)) ^ ") {\n" 
			^ (tr_block c scopes cur)
			^ "}\n}\n"  
		    )
		in
		write super_str ("Func_" ^ str ^ ".java");
		"Func_" ^ str ^ " " ^ str ^ " = new Func_" ^ str ^ "()"
		| _ ->
		    "THIS SHOULD NEVER BE CALLED EITHER"
		(* function in a figure. this should never be called *)
		(*
		(tr_typ a) ^ " " ^ str ^ "(" 
		^ (tr_list2 (tr_arglist b)) ^
		") {\n" ^ (tr_block c scopes cur) ^ "}\n"
		) *))
	    | _ -> "" ^ (tr_typ_opt typ_opt str scopes cur) ^ 
		" " ^ id_prefix ^ str ^ (tr_expr_opt expr_opt scopes cur ) 
	) 
		) 
	| Param -> "private " ^ (tr_typ_opt typ_opt str scopes cur) 
	    ^ " " ^ id_prefix ^ str ^ (tr_expr_opt expr_opt scopes cur)
	| Comp -> "private " ^ (tr_typ_opt typ_opt str scopes cur) 
	    ^ " " ^ id_prefix ^ str ^ (tr_expr_opt expr_opt scopes cur)
	)
		
		
		)
	
		
    in 
	let almost = (List.map (fun x -> tr_decl x scopes cur ) decls)
	in tr_list almost
	    
in

(* ------------------------ FIGURE TRANSLATION ------------------------ *)

let tr_figdecls decls scopes cur = 

    let tr_typ_opt arg str scopes cur = match arg with
	  None -> let containingScope = (try List.nth scopes cur with
	    Invalid_argument(str) -> raise (Internal_error ("scopes not valid in tr_typ_opt"))) 
	  (* this is where you search teh scopes list for the type*) 
	    in
	    tr_typ (StringMap.find str containingScope.symbols) 
	| Some(typ) -> tr_typ typ
    in

    let tr_expr_opt arg scopes cur = match arg with
	  None -> ""
	| Some(expr) -> " = " ^ (tr_expr expr scopes cur)

    in

    let tmp_init = ref "" in
    let tmp_comp = ref "" in
    let tmp_binit = ref "" in

    let tr_figdecl decl scopes cur = (match decl with
	(decl_typ, str, typ_opt, expr_opt) ->

	if is_captured (str, cur) then
	    let other = tr_expr_opt expr_opt scopes cur in
	    if other = "" then "" else
		(
		(*
		   let mask = tr_expr_opt expr_opt scopes cur in
		   let mask2 = String.sub mask 2 ((String.length mask) -2)
		   in
		   tmp_init := !tmp_init ^ 
		   (captures_helper (str, cur) ^
		   " = (AbstractType) (" ^ mask2 ^ ")") ^ ";\n";
		   ""
		   *)
		    tmp_binit := !tmp_binit ^ 
		   (captures_helper (str, cur) ^
		   (tr_expr_opt expr_opt scopes cur)) ^ ";\n";
		   ""

		)
	else

	(

	match decl_typ with 
    	  Var ->   
	    (match expr_opt with
	      None ->  "" ^ (tr_typ_opt typ_opt str scopes cur) ^ " " 
		^ id_prefix ^ str ^ (tr_expr_opt expr_opt scopes cur) 
	    | Some(expr) ->   
		(match expr with
		  Function(a, b, c) -> 
		    let tr_arglist arglist = 
			let tr_s x = match x with
			(str, typ) -> (tr_typ typ) ^ " " ^ str
		    in
			List.map (fun x -> tr_s x) arglist   
		    (* in
		    function in a figure
		    let func = (*(tr_typ a) ^ " " ^ str ^ "(" ^ 
		    (tr_list2 (tr_arglist b)) ^ ") {\n" ^*) 
		    (tr_block c scopes cur) (*^ "}\n"*)*)

		    in let func_name = 
		"Func_" ^ str ^ (string_of_int cur) 

		    in let func_cls = header ^ "class " ^ func_name ^ 

			(if (str = "_update") then 
			    " extends UpdateFunctionType {\n" else
				" extends FunctionType {\n" )
			
			^ "public " ^ (tr_typ a) 
			^ " call(" ^ 
			(tr_list2 (tr_arglist b)) ^ ") {\n" 
			^ (tr_block c scopes cur)
			^ "}\n}\n"  
		    in
		    write func_cls (func_name ^ ".java");

		    let superstr = 
		    (if (str = "_update") then 
			"_update = (UpdateFunctionType) (new " 
			^ func_name ^ "())"
		    else
			"public " ^ func_name ^ " " ^ str ^ " = new " ^ func_name ^ 
			"()")
		    in
		    tmp_binit := !tmp_binit ^ superstr ^ ";\n";
		    ""
		    
		| _ -> "" ^ (tr_typ_opt typ_opt str scopes cur) ^ 
		    " " ^ id_prefix ^ str ^ (tr_expr_opt expr_opt scopes cur ) 
		) 
	    ) 
	| Param -> 
	    (
	    match (expr_opt)  with
	      None -> "" ^ (tr_typ_opt typ_opt str scopes cur) 
		^ " " ^ id_prefix ^ str ^ 
		(tr_expr_opt expr_opt scopes cur)
	    | Some(expr) -> 
		(match expr with
		  Function(a, b, c) -> 
		    let tr_arglist arglist = 
			let tr_s x = match x with
			(str, typ) -> (tr_typ typ) ^ " " ^ str
		    in
			List.map (fun x -> tr_s x) arglist   
		    (* in
		    function in a figure
		    let func = (*(tr_typ a) ^ " " ^ str ^ "(" ^ 
		    (tr_list2 (tr_arglist b)) ^ ") {\n" ^*) 
		    (tr_block c scopes cur) (*^ "}\n"*)*)

		    in let func_name = 
			"Func_" ^ str ^ (string_of_int cur) 

		    in let func_cls = header ^ "class " ^ func_name ^ 
			" extends FunctionType {\n" 
			^ "public " ^ (tr_typ a) 
			^ " call(" ^ 
			(tr_list2 (tr_arglist b)) ^ ") {\n" 
			^ (tr_block c scopes cur)
			^ "}\n}\n"  
		    in
		    write func_cls (func_name ^ ".java");

		    let superstr = 
		    ("" ^ "FunctionType" ^
			" " ^ str ^ " = new " ^ func_name ^ 
			"()")
		    in
		  (*  tmp_init := !tmp_init ^ superstr ^ ";\n";
		  *)
		    superstr
		   
		| _ -> "" ^ (tr_typ_opt typ_opt str scopes cur) 
		^ " " ^ id_prefix ^ str ^ 
		(tr_expr_opt expr_opt scopes cur)
		)
	    )
	| Comp -> 
	    (* add str to components list *)
	    tmp_comp := (!tmp_comp ^ "components.add(" ^ (str) ^ ");\n");
	    (* add the following to comp_init *)
	    tmp_init := (!tmp_init ^ str
			^(tr_expr_opt expr_opt scopes cur)^";\n");
	    (* return the following as the code *)
	    ("" ^ (tr_typ_opt typ_opt str scopes cur) 
	    ^ " " ^ id_prefix ^ str) 
		
		)
	)
		
    in 
	let almost = (List.map (fun x -> tr_figdecl x scopes cur) decls)
	in 
	
	(tr_list almost) ^ ("void first_init() {\n" ^ !tmp_comp ^ "}\n")
	^ ("void create() {\n" ^ !tmp_init ^ "}\n" ^ 
	"void b_init() {\n"^ !tmp_binit ^ "}\n")
	
in

(* translates a figure into java and produces a .java file *)
let tr_figure name decls scopes cur captures =

    (* -------------------- CREATE THE CONSTRUCTOR ------------------ *) 
    let create_constructor name decls scopes cur = 
	
	(* helper function will return a strmap of param -> int index *)
	let helper decl lst = 
	    match decl with
	    (decl_typ, str, typ_opt, expr_opt) ->
		(
		    match decl_typ with
		      Param -> 
			str::lst 
		    | _ -> lst 
		)

	in
	
	let counter = ref (-1) in
	
	let giant_lst = 
	    List.fold_left (fun a b -> helper b a) [] decls
	in
	figure_record := StringMap.add name giant_lst !figure_record;
	let giant_stmt = List.fold_left 
	    (fun a b -> 
		let scope_b = scope_of_symbol b cur scopes in
		counter := !counter + 1;
		a ^ "if (arg[" ^ (string_of_int !counter)
		^ "] != null) { " ^ 
		(if (is_captured (b, scope_b)) then 
		(captures_helper (b, scope_b))
		else b)
		^ " = (" 
		^ (tr_typ (type_of_symbol b cur scopes)) ^ ") arg[" ^ 
		(string_of_int !counter)^ "]; }\n" 
	    ) 
	    "" (List.rev giant_lst)
	in
	"public " ^ name ^ "(AbstractType... arg) {\nb_init();\n" ^
	giant_stmt ^ 
	"create();\nfirst_init();\nif (_update != null)\n_animating =true;\n}"	
    in

    let constructor = create_constructor name decls scopes cur in

    let strstr = 
	header ^ "public class " ^ name ^ " extends Figure {\n" ^
	(tr_figdecls decls scopes cur) ^ constructor ^ "\n}\n"
	
    in 
    write strstr (name ^ ".java"); 
	
    "//Class " ^ name ^ " created."

in
(* stdlib translation *)
    let tr_stdlib_stmt stmt =  
	let tr_stdlib_decl decl = 
	match decl with (decl_typ , str , typ_opt , expr_opt) ->
	    (match expr_opt with
		  None -> 
		    let typ = (match (typ_opt) with 
			  None -> "Very big problem."
			| Some(typ) -> tr_typ typ)
		    in
		    if (is_captured (str, 0)) then
			(captures_helper (str, 0)) ^ " = Stdlib.Lib." ^ str ^
			    ";\n"
		    else
			(stdlib := str::!stdlib;
			typ ^ " " ^  str ^ " = Stdlib.Lib." ^ str ^ ";\n"
			)
		| Some(expr) -> "Big problem as well."
	    )

	in (match stmt with
	      Decl(decls) -> List.fold_left 
		(fun x y -> x ^ tr_stdlib_decl y) "" decls
	    | _ -> "BIG PROBLEM"
	    )
	
in
    match stmt with
      Expr(expr) -> tr_expr expr scopes cur
    | Figure(arg1, arg2, decls) -> tr_figure arg2 decls scopes arg1 cur
    | Decl(arg) -> 
	if (!stdlib_parse) then
		(stdlib_parse := false;
		tr_stdlib_stmt stmt)
	    else tr_decls arg scopes cur 
    | Assign(arg1, arg2) -> 
	if ( ((type_of_expr arg1 cur scopes) = "FloatType") &&
		((type_of_expr arg2 cur scopes) = "IntType"))
	then
	    (tr_expr arg1 scopes cur) ^ 
	    " = new FloatType(" ^ (tr_expr arg2 scopes cur)^ ")"
	else  
	
	(if ( (type_of_expr arg2 cur scopes) = "FunctionType") 

	then
	    (match arg2 with 
	    Function(a,b,c) ->
		let func_name =
		    arbitrary_function_counter :=
			!arbitrary_function_counter + 1;
		    "Func_" ^ (string_of_int cur) ^ "_" ^
			(string_of_int !arbitrary_function_counter)
		in
		let tr_arglist arglist = 
		    let tr_s x = match x with
		    (str, typ) -> (tr_typ typ) ^ " " ^ str
		in
		    List.map (fun x -> tr_s x) arglist   
		    in
		let super_str = 
		    (header ^"public class " ^
			func_name ^ " extends FunctionType {\n" 
			^ (tr_typ a) ^ " call(" ^ 
			(tr_list2 (tr_arglist b)) ^ ") {\n" 
			^ (tr_block c scopes cur)
			^ "}\n}\n"  
		    )
		in
		write super_str (func_name ^ ".java");
		" new " ^ func_name ^ "()"
		
	    | _ -> "This is purely a physics problem.")
	else
	(if ( ((type_of_expr arg1 cur scopes)) = "Subscript") 
	    then 
	    (match arg1 with
		  Subscript(a1, a2) -> 
		    "(" ^ (tr_expr a1 scopes cur) 
		    ^ ").set(" ^(tr_expr a2 scopes cur) ^", "^
		    (tr_expr arg2 scopes cur ) ^ ")"
	
		| _ -> "Another pure physics problem.")

	else  
		
		
	(if (type_of_expr2 arg1 cur scopes)="Id" then 
	(match arg1 with
	    Id(a1) -> 
		(* is_captured arg1, cur is not returning true *)
		let arg_scope_index =
		    scope_of_symbol a1 cur scopes
		in
		if arg_scope_index = 9001 
		    then (print_endline "OVER 9000 ERROR";"") else		
		if (is_captured (a1, arg_scope_index)) then
		    captures_helper (a1, arg_scope_index)  
			 ^ " = " ^
			(tr_expr arg2 scopes cur)


		else 
		    if ((in_stdlib a1)&&(arg_scope_index = 0)) then
			("Stdlib.Lib." ^ a1) ^ " = " ^
			(tr_expr arg2 scopes cur)

		    else
			(tr_expr arg1 scopes cur) ^ " = " ^
			(tr_expr arg2 scopes cur)

	    | _ -> "too many laws of physics were broken")

	else (tr_expr arg1 scopes cur) ^ " = " ^
	    (tr_expr arg2 scopes cur)
	    )))
    | Return(arg) -> 
	    (match arg with
		  None -> "return;"
	    	| Some(ex) -> 

	(if ((type_of_expr ex cur scopes) = "FunctionType") 

	then
	    (match ex with Function(a,b,c) ->
		let func_name =
		    arbitrary_function_counter :=
			!arbitrary_function_counter + 1;
		    "Func_" ^ (string_of_int cur) ^ "_" ^
			(string_of_int !arbitrary_function_counter)
		in
		let tr_arglist arglist = 
		    let tr_s x = match x with
		    (str, typ) -> (tr_typ typ) ^ " " ^ str
		in
		    List.map (fun x -> tr_s x) arglist   
		    in
		let super_str = 
		    (header ^"public class " ^
			func_name ^ " extends FunctionType {\n" 
			^ (tr_typ a) ^ " call(" ^ 
			(tr_list2 (tr_arglist b)) ^ ") {\n" 
			^ (tr_block c scopes cur)
			^ "}\n}\n"  
		    )
		in
		write super_str (func_name ^ ".java");
		"return new " ^ func_name ^ "()"
	    | _ -> "THIS is purely a physic problem 2")
	else
	    "return " ^ (tr_expr ex scopes cur) ^ ""))
    | If(arg1, arg2, arg3) -> "if ((" ^ 
	(tr_expr arg1 scopes cur) ^ ").value) {\n" ^ 
	(tr_block arg2 scopes cur) ^ "\n} else {\n" ^ 
	(tr_block arg3 scopes cur) ^ "\n } \n"
    | While(arg1, arg2) -> "while ((" ^
	(tr_expr arg1 scopes cur) ^").value) {\n" 
	^ (tr_block arg2 scopes cur) ^ "\n}\n"
    | DoWhile(arg1, arg2)  -> "do {\n" ^ 
	(tr_block arg1 scopes cur) ^ "\n} while ((" 
	^ (tr_expr arg2 scopes cur) ^ ").value)\n" 

(* no bindings work as of now =(  *)

(* -------------------------- BINDINGS --------------------------------*)

let analyze_bindings_decl decl s_num bindings_list = 
    match decl with
	(decl_typ, str, typ_opt, expr_opt) -> "" (* TODO: BINDINGS LIST *)

let analyze_bindings_decls decls s_num bindings_list =
    List.fold_left (fun a b -> analyze_bindings_decl b s_num a) 
	bindings_list decls

let analyze_bindings_stmt stmt scopes bindings_list = 
    match stmt with
	  Figure(num,str,decls) ->  bindings_list (* TODO: BINDINGS LIST *)
	| _ -> bindings_list

let create_bindings_class program scopes = ""

(* -------------------------- CAPTURES --------------------------------*)

(* add captures to the list *)
let analyze_captures_in_scope scope captures_list = 
    let helper = (fun a b d -> 
		(match d with
		    (a2, b2) -> 
			(a2 = a) && (b2 = b)))
    in
    match scope.capture with
	  None -> 
	    captures_list
	| Some(strmap) -> 
	    StringMap.fold 
	    (fun a b c -> 
		if (List.exists (helper a b) c) 
		    then c 
		    else ((a,b)::c)) 
	    strmap captures_list

let analyze_captures_in_scopes scopes captures =
    List.fold_left (fun a b -> (analyze_captures_in_scope b a) ) captures scopes

(* format captures for debug printing *)
let of_captures captures =
    let print_capture x = match x with ((a,b), c) ->
	( "[(" ^ a ^ ", " ^ 
	    (string_of_int b) ^ ") -> " ^ (string_of_int c) ^ "]")
    in
    let to_print = "Captures: " ^ List.fold_left (fun x c -> 
	x ^(if x="" then "" else ", ") ^(print_capture c)) 
	"" captures
    in to_print

   
let get_captures scopes = 
    let all_the_captures = analyze_captures_in_scopes scopes []
    in
    let index = ref (-1) in 

    let result = (List.fold_left 
	(fun lst c -> 
	    index := !index + 1; 
	    (c,!index)::lst) 
	[] all_the_captures)
    in
(*    (print_endline (of_captures result));
*)    result
(* each 'capture' is a ((variable name, scope), index) *)


(* translate a program into java *)
let tr_program program scopes current = 
    captures := get_captures scopes;
    let almost = List.map (fun x -> tr_stmt x scopes current) program
    in 
    
    let almost_main_fun = 
	header ^ "public abstract class GeneratedMain {\n" ^
	    "public static void generatedMain() {" ^
	    "Captures.record = new AbstractType[" 
		^ (string_of_int(List.length !captures)) ^ "];" ^
	    tr_list almost ^
	    "}\n}\n"
    in 
	write almost_main_fun "GeneratedMain.java"
